local atrUF = CreateFrame("Frame", "atrUF", UIParent)
atrUF:SetAllPoints(UIParent)
_G["atrUF"] = atrUF

atrUF.DEFAULT_STYLE = 1
atrUF.MINI_STYLE = 2
atrUF.PARTY_STYLE = 3

atrUF.units = {}

atrUF.auraSettings = {
    default = {
        buffs = {
            size = 20,
            num = 40,
            anchor = "BOTTOMLEFT",
            xdir = 1,
            ydir = 1,
            cols = 10,
            rows = 4, -- num/cols
            filter = "HELPFUL",
        },
        debuffs = {
            size = 20,
            num = 40,
            anchor = "TOPLEFT",
            xdir = 1,
            ydir = -1,
            cols = 10,
            rows = 4, -- num/cols
            filter = "HARMFUL",
        }
    },
    mini = {
        buffs = {
            size = 20,
            num = 6,
            anchor = "BOTTOMRIGHT",
            xdir = -1,
            ydir = 1,
            cols = 3,
            rows = 2, -- num/rows
            filter = "HELPFUL",
        },
        debuffs = {
            size = 20,
            num = 4,
            anchor = "BOTTOMLEFT",
            xdir = 1,
            ydir = 1,
            cols = 2,
            rows = 2, -- num/rows
            filter = "HARMFUL",
        }
    },
    party = {
        buffs = {
            size = 14,
            num = 2,
            anchor = "TOPLEFT", -- Not used
            xdir = 1,
            ydir = 1,
            cols = 2,
            rows = 1, -- num/rows
            filter = "HELPFUL|RAID",
        },
        debuffs = {
            size = 14,
            num = 2,
            anchor = "TOPRIGHT", -- Not used
            xdir = -1,
            ydir = 1,
            cols = 2,
            rows = 1, -- num/rows
            filter = "HARMFUL",
        }
    }
}

-------------------------

function atrUF:UpdateAuraSet(unit, debuffSet)
    local auraSet = debuffSet and self[unit].debuffs or self[unit].buffs
    local i = 1
    
    while(true) do
        local name, rank, icon, numberCount, debuffType, duration, expires, caster, stealable = UnitAura(unit, i, auraSet.settings.filter)
        local button = auraSet.auras[i]
        if(not name or not button) then break end
        
        button:Show()
        button.index = i
        button.expires = expires
        button.texture:SetTexture(icon)
        
        if(numberCount > 1) then
            button.count:Show()
            button.count:SetText(numberCount)
        else
            button.count:Hide()
        end
        
        if(debuffSet and debuffType) then
            color = DebuffTypeColor[debuffType]
        else
            color = {r = 0, g = 0, b = 0}
        end
        button.overlay:SetBackdropColor(color.r, color.g, color.b)

        
        if(atrUF[unit].style ~= atrUF.PARTY_STYLE) then
            self:UpdateAuraTimer(auraSet, i)
        end
            
        i = i +1
    end
    
    auraSet.numVisible = i
    
    for i = i, #auraSet.auras do
        auraSet.auras[i]:Hide()
    end
end

function atrUF:UpdateAuras(unit)
    self:UpdateAuraSet(unit, false)
    self:UpdateAuraSet(unit, true)
end

function atrUF:UpdateAuraTimer(auraSet, index)
    local aura = auraSet.auras[index]
    if(not aura or not aura.index) then return end
    local timer = aura.timer
    local time = GetTime()
    local n, _, _, _, _, duration, expires, caster, _ = UnitAura(auraSet:GetParent().unit, aura.index, auraSet.settings.filter)

    if(expires) then
        local start = expires-duration
        local elapsed = time-start
        local remainFormat, remainValue = SecondsToTimeAbbrev(math.floor(duration-elapsed))
        local _, _, remainUnit, _ = string.find(remainFormat, "(%l)$")
        
        local format
        
        if(remainUnit == "h") then
            format = "|cffffff00%s|r"
        elseif(remainUnit == "m") then
            format = "|cff00ff00%s|r"
        elseif(remainUnit == "s" and remainValue <= 3) then
            format = "|cffff0000%s|r"
        else
            format = "|cffffffff%s|r"
        end
        
        if(remainValue <= 0) then
            remainValue = ""
        end
        
        timer.text:SetFormattedText(format, remainValue)
    end
end

function atrUF:UpdateName(unit)
    if(not UnitExists(unit)) then return end
    local color
    local class = select(2, UnitClass(unit))
    
    if(class) then
        color = atrCore.colors[class]
    else
        color = {1, 1, 1}
    end
    
    self[unit].name:SetFormattedText(atrCore:TextHexColor(color[1], color[2], color[3], "%s"), UnitName(unit))
end

function atrUF:UpdateLevel(unit)
    local f = self[unit]
    if(not f) then return end
    
    if(f.level) then
        local level = UnitLevel(unit)
        if (level < 0) then level = "" end
        local classificationTypes = {
            ["worldboss"] = "B+",
            ["rareelite"] = "R+",
            ["elite"] = "+",
            ["rare"] = "R",
            ["normal"] = "",
            ["trivial"] = "",
        }
        local classification = ""
        local color = {1, 1, 1}
        
        if(not UnitIsPlayer(unit)) then
            color[1], color[2], color[3] = UnitSelectionColor(unit)
        else
            color = atrCore.colors[select(2, UnitClass(unit))]
        end
    
        classification = classificationTypes[UnitClassification(unit)]
        f.level:SetFormattedText(atrCore:TextHexColor(color[1], color[2], color[3], "%s%s"), level, classification)
    end
end

function atrUF:UpdateHealth(unit)
    local f = self[unit]    
    local current, max = UnitHealth(unit), UnitHealthMax(unit)
    local percent, value = (current/max)*100
    
    if(UnitIsDead(unit)) then
        value = DEAD
    elseif(UnitIsPlayer(unit) and not UnitIsConnected(unit)) then
        value = PLAYER_OFFLINE
        percent = 0
    else
        if(current >= 1000000) then
            value = atrCore:Round(current/1000000, 1).."m"
        elseif(current >= 1000) then
            value = atrCore:Round(current/1000, 1).."k"
        else
            value = current
        end
    end
    
    f.health:SetValue(percent)
    f.health.value:SetText(value)
end

function atrUF:UpdatePower(unit)
    local f = self[unit]
    local current, max = UnitPower(unit), UnitPowerMax(unit)
    local percent = (max == 0) and 0 or (current/max)*100
    local type = select(2, UnitPowerType(unit))
    local offline = UnitIsPlayer(unit) and not UnitIsConnected(unit)
    local color = atrCore.colors[type]
    if(not color) then color = atrCore.colors.MANA end
    
    if(offline or max == 0) then percent = 0 end
    
    f.power:SetValue(percent)
    f.power:SetStatusBarColor(color[1], color[2], color[3])
    f.power:SetBackdropColor(color[1]*.3, color[2]*.3, color[3]*.3)
        
    if(f.power.value) then
        local value
        
        if(offline or current == 0) then
            value = ""
        elseif(current >= 1000000) then
            value = atrCore:Round(current/1000000, 1).."m"
        elseif(current >= 1000) then
            value = atrCore:Round(current/1000, 1).."k"
        else
            value = current
        end
        
        f.power.value:SetText(value)
    end
end

function atrUF:UpdateThreat(unit)
    local color
    local unittarget = unit.."target"
    
    if(UnitExists(unittarget)) then
        local status = UnitThreatSituation(unit, unittarget)
        color = status and atrCore.colors.threat[status] or atrCore.colors.threat.none
    else
        color = atrCore.colors.threat.none
    end
    
    self[unit].health:SetStatusBarColor(unpack(color))
end

function atrUF:UpdateRaidIcon(unit)
    local index = GetRaidTargetIndex(unit)
    local f = self[unit]
    
    if(index) then
        f.raidIcon:Show()
        SetRaidTargetIconTexture(f.raidIcon.texture, index)
    else
        f.raidIcon:Hide()
    end
end

function atrUF:UpdateRange(unit)
    local fade = true
    if(UnitInParty(unit) or UnitInRaid(unit)) then
        if(UnitIsConnected(unit) and UnitInRange(unit)) then
            fade = false
        end
    else
        fade = false
    end
    
    self[unit]:SetAlpha(fade and .5 or 1)
end

function atrUF:SpawnAuraSet(frame, isDebuff)    
    local settings = isDebuff and frame.auraSettings.debuffs or frame.auraSettings.buffs
    local auraSet = CreateFrame("Frame", nil, frame)
    if(isDebuff) then
        frame.debuffs = auraSet
    else
        frame.buffs = auraSet
    end
    auraSet.isDebuffSet = isDebuff
    auraSet.settings = settings
    auraSet.auras = {}
    auraSet:SetWidth(settings.cols*settings.size)
    auraSet:SetHeight(settings.rows*settings.size)
    if(frame.style ~= atrUF.PARTY_STYLE) then
        if(settings.anchor == "BOTTOMRIGHT") then
            auraSet:SetPoint(settings.anchor, frame.health, "TOPRIGHT", 0, 5)
        elseif(settings.anchor == "BOTTOMLEFT") then
            auraSet:SetPoint(settings.anchor, frame.health, "TOPLEFT", 0, 5)
        elseif(settings.anchor == "TOPLEFT") then
            auraSet:SetPoint(settings.anchor, frame.health, "BOTTOMLEFT", 0, -10)
        elseif(settings.anchor == "TOPRIGHT") then
            auraSet:SetPoint(settings.anchor, frame.health, "BOTTOMRIGHT", 0, -10)
        end
    else
        if(not isDebuff) then
            auraSet:SetPoint("TOPLEFT", frame.health, "TOPLEFT", settings.size/2, settings.size/2)
        else
            auraSet:SetPoint("TOPRIGHT", frame.health, "TOPRIGHT", -settings.size/2, settings.size/2)
        end
    end
    
    local i
    local col, row = 0, 0
    for i = 1, settings.num do
        if(col >= settings.cols) then
            col = 0
            row = row+1
        end
        
        atrUF:SpawnAura(frame, isDebuff, col, row)
        
        col = col+1
    end
    
    if(frame.style ~= atrUF.PARTY_STYLE) then
        auraSet.nextUpdate = 0
        auraSet:SetScript("OnUpdate", function(s, elapsed)
            if(s.nextUpdate <= 0) then
                s.nextUpdate = 1
                
                local i
                for i = 1, s.numVisible do
                    atrUF:UpdateAuraTimer(s, i)
                end
            else
                s.nextUpdate = s.nextUpdate-elapsed
            end
        end)
    end
end

function atrUF:SpawnAura(frame, isDebuff, col, row)
    local auraSet = isDebuff and frame.debuffs or frame.buffs
    local settings = isDebuff and frame.auraSettings.debuffs or frame.auraSettings.buffs
    local ab = CreateFrame("Button", nil, auraSet)
    
    table.insert(auraSet.auras, ab)
            
    ab:SetWidth(settings.size)
    ab:SetHeight(settings.size)
    ab:SetPoint(settings.anchor, auraSet, settings.anchor, col*settings.size*settings.xdir, row*settings.size*settings.ydir)

    ab.texture = ab:CreateTexture()
    ab.texture:SetAllPoints(ab)
    ab.texture:SetTexCoord(0, 1, 0, 1)

    ab.overlay = CreateFrame("Frame", nil, ab)
    ab.overlay:SetAllPoints(ab)
    ab.overlay:SetBackdrop(atrCore.art.backdrop.aura)
    ab.overlay:SetBackdropColor(unpack(atrCore.art.backdrop.aura.color))

    if(frame.style ~= atrUF.PARTY_STYLE) then
        ab.timer = CreateFrame("Frame", nil, ab)
        ab.timer.text = ab.timer:CreateFontString(nil, "OVERLAY")
        ab.timer.text:SetPoint("CENTER", ab, "CENTER", .5, 0)
        ab.timer.text:SetFont(unpack(atrCore.art.font.aura.timer))
        ab.timer.text:SetJustifyH("CENTER")
        ab.timer.text:SetJustifyV("MIDDLE")
        ab.timer.text:SetText("99")
    end

    ab.count = ab.overlay:CreateFontString(nil, "OVERLAY")
    ab.count:SetPoint("CENTER", ab, "BOTTOMRIGHT", -2, 2)
    ab.count:SetFont(unpack(atrCore.art.font.aura.count))
    ab.count:SetTextColor(1, 1, 1)
    ab.count:SetShadowOffset(1, -2)
    ab.count:SetJustifyH("CENTER")
    ab.count:SetJustifyV("MIDDLE")

    ab:SetScript("OnHide", function(self)
        self.started = nil
        self.duration = nil
    end)
    ab:SetScript("OnEnter", function(self)
        local auraSet = self:GetParent()
        local unitFrame = auraSet:GetParent()
        local filter
        
        if(auraSet.isDebuffSet) then
            filter = unitFrame.auraSettings.debuffs.filter
        else
            filter = unitFrame.auraSettings.buffs.filter
        end
        
        GameTooltip:SetOwner(self, "ANCHOR_BOTTOMRIGHT")
        GameTooltip:SetUnitAura(unitFrame.unit, self.index, filter)
    end)
    ab:SetScript("OnLeave", function(self)
        GameTooltip:Hide()
    end)
    if(not isDebuff and (frame.unit == "player" or frame.unit == "pet")) then
        ab:RegisterForClicks("RightButtonUp")
        ab:SetScript("OnClick", function(self)
            local auraSet = self:GetParent()
            local unitFrame = auraSet:GetParent()
            local filter
            
            if(auraSet.isDebuffSet) then
                filter = unitFrame.auraSettings.debuffs.filter
            else
                filter = unitFrame.auraSettings.buffs.filter
            end
            
            CancelUnitBuff(unitFrame.unit, self.index, filter)
        end)
    end

    return ab
end

function atrUF:NilFunction() end
function atrUF:HideDefaultUnit(unit)
    local blank = self.NilFunction
    
    if(unit == "player") then
		PlayerFrame:UnregisterAllEvents()
		PlayerFrame.Show = blank
		PlayerFrame:Hide()
        
        PlayerFrameHealthBar:UnregisterAllEvents()
		PlayerFrameManaBar:UnregisterAllEvents()
    elseif(unit == "pet") then
        PetFrame:UnregisterAllEvents()
		PetFrame.Show = blank
		PetFrame:Hide()

		PetFrameHealthBar:UnregisterAllEvents()
		PetFrameManaBar:UnregisterAllEvents()
    elseif(unit == "target") then
        TargetFrame:UnregisterAllEvents()
		TargetFrame.Show = blank
		TargetFrame:Hide()

		TargetFrameHealthBar:UnregisterAllEvents()
		TargetFrameManaBar:UnregisterAllEvents()
		TargetFrameSpellBar:UnregisterAllEvents()

		ComboFrame:UnregisterAllEvents()
		ComboFrame.Show = blank
		ComboFrame:Hide()
    elseif(unit == "targettarget") then
        TargetofTargetFrame:UnregisterAllEvents()
        TargetofTargetFrame.Show = blank
        TargetofTargetFrame:Hide()

        TargetofTargetHealthBar:UnregisterAllEvents()
        TargetofTargetManaBar:UnregisterAllEvents()
    elseif(unit == "focus") then
        FocusFrame:UnregisterAllEvents()
		FocusFrame.Show = blank
		FocusFrame:Hide()

		FocusFrameHealthBar:UnregisterAllEvents()
		FocusFrameManaBar:UnregisterAllEvents()
		FocusFrameSpellBar:UnregisterAllEvents()
    elseif(unit == "focustarget") then
        TargetofFocusFrame:UnregisterAllEvents()
        TargetofFocusFrame.Show = blank
        TargetofFocusFrame:Hide()
        
        TargetofFocusHealthBar:UnregisterAllEvents()
        TargetofFocusManaBar:UnregisterAllEvents()
    else
        local partyIndex = select(3, string.find(unit, "party([0-9]?)"))

        if(partyIndex) then
            local f = _G["PartyMemberFrame"..partyIndex]
            if(f) then
                f:UnregisterAllEvents()
                f.Show = blank
                f:Hide()
            end
        end
    end
end

function atrUF:Spawn(unit, style)    
    local style = style and style or atrUF.DEFAULT_STYLE
    local isPartyFrame = string.find(unit, "party[0-9]?")
    local isRaidFrame = string.find(unit, "raid[0-9]?")
    table.insert(self.units, unit)
    
    local f = CreateFrame("Button", "atrUF_"..unit, self, "SecureUnitButtonTemplate")
    self[unit] = f
    self:HideDefaultUnit(unit)

    f:SetFrameStrata("BACKGROUND")
    f.unit = unit
    f.style = style
    f.isPartyFrame = isPartyFrame
    f.isRaidFrame = isRaidFrame
    f.menu = function(cf)
        local unit = cf.unit:gsub('(.)', string.upper, 1)
        if(_G[unit..'FrameDropDown']) then
            ToggleDropDownMenu(1, nil, _G[unit..'FrameDropDown'], 'cursor')
        end
    end
    f:RegisterForClicks("AnyUp")
    f:SetAttribute("unit", unit)
    f:SetAttribute("type1", "target")
    f:SetAttribute("type2", "menu")
    RegisterUnitWatch(f)
    f:Hide()

    if(style == atrUF.MINI_STYLE) then
        f:SetWidth(110)
        f:SetHeight(40)
        f.auraSettings = atrUF.auraSettings.mini
    elseif(style == atrUF.PARTY_STYLE) then
        f:SetWidth(100)
        f:SetHeight(40)
        f.auraSettings = atrUF.auraSettings.party
    else
        f:SetWidth(210)
        f:SetHeight(40)
        f.auraSettings = atrUF.auraSettings.default
    end

    f:SetBackdrop(atrCore.art.backdrop.default)
    f:SetBackdropColor(unpack(atrCore.art.backdrop.default.color))
    f:SetBackdropBorderColor(unpack(atrCore.art.backdrop.default.borderColor))
        
    f.health = CreateFrame("StatusBar", nil, f)
    f.health:SetStatusBarTexture("Interface\\AddOns\\atrCore\\foreground_healthbar")
    f.health:SetStatusBarColor(unpack(atrCore.colors.threat.none))
    f.health:SetBackdrop({
        bgFile = "Interface\\AddOns\\atrCore\\background",
        tile = true
    })
    f.health:SetBackdropColor(unpack(atrCore.art.backdrop.default.backdropColor))
    f.health:SetPoint("TOPLEFT", 5, -5)
    f.health:SetPoint("BOTTOMRIGHT", -5, 10)
    f.health:SetMinMaxValues(0, 100)
    
    f.health.value = f.health:CreateFontString(nil, "OVERLAY")
    f.health.value:SetFont(unpack(atrCore.art.font.default))
	f.health.value:SetPoint("RIGHT", -5, 0)
	f.health.value:SetTextColor(1, 1, 1)
	f.health.value:SetJustifyH("RIGHT")
	f.health.value:SetShadowOffset(1, -1)
    f.health.value:SetText(1234)
    
    f.power = CreateFrame("StatusBar", nil, f)
    f.power:SetStatusBarTexture("Interface\\AddOns\\atrCore\\foreground_powerbar")
    f.power:SetBackdrop({
        bgFile = "Interface\\AddOns\\atrCore\\background",
        tile = true
    })
    f.power:SetBackdropColor(1, 0, 0, 1)
    f.power:SetPoint("TOPLEFT", f.health, "BOTTOMLEFT")
    f.power:SetPoint("BOTTOMRIGHT", -5, 5)
    f.power:SetMinMaxValues(0, 100)
    
    if(style ~= atrUF.MINI_STYLE and style ~= atrUF.PARTY_STYLE) then
        f.power.value = f.health:CreateFontString(nil, "OVERLAY")
        f.power.value:SetFont(unpack(atrCore.art.font.default))
        f.power.value:SetPoint("RIGHT", f.health.value, "LEFT", -10, 0)
        f.power.value:SetTextColor(1, 1, 1)
        f.power.value:SetJustifyH("RIGHT")
        f.power.value:SetShadowOffset(1, -1)
        f.power.value:SetText(5678)
    end
    
    f.name = f.health:CreateFontString(nil, "OVERLAY")
    f.name:SetFont(unpack(atrCore.art.font.default))
    f.name:SetPoint("LEFT", 5, 0)
    if(f.power.value) then
        f.name:SetPoint("RIGHT", f.power.value, "LEFT", -10, 0)
    else
        f.name:SetPoint("RIGHT", f.health.value, "LEFT", -10, 0)
    end
    f.name:SetTextColor(1, 1, 1)
    f.name:SetJustifyH("LEFT")
    f.name:SetShadowOffset(1, -1)
    f.name:SetText("Name")
    
    if(unit == "target") then
        f.level = f.health:CreateFontString(nil, "OVERLAY")
        f.level:SetFont(unpack(atrCore.art.font.default))
        f.level:SetPoint("LEFT", 5, 0)
        f.level:SetTextColor(1, 1, 1)
        f.level:SetJustifyH("LEFT")
        f.level:SetShadowOffset(1, -1)
        f.level:SetText("80")
        
        f.name:SetPoint("LEFT", f.level, "RIGHT", 5, 0)
    end
    
    f.raidIcon = CreateFrame("Frame", nil, f)
    f.raidIcon:SetFrameStrata("HIGH")
    f.raidIcon:SetWidth(f.auraSettings.buffs.size*.9)
    f.raidIcon:SetHeight(f.auraSettings.buffs.size*.9)
    f.raidIcon:SetPoint("CENTER", f, "TOP", 0, -f.raidIcon:GetHeight()/4)
    f.raidIcon:Hide()
    
    f.raidIcon.texture = f.raidIcon:CreateTexture()
    f.raidIcon.texture:SetAllPoints(f.raidIcon)
    f.raidIcon.texture:SetTexture("Interface\\TargetingFrame\\UI-RaidTargetingIcons")
    
    atrUF:SpawnAuraSet(f, false)
    atrUF:SpawnAuraSet(f, true)
    
    ------------------------------------
    if(unit == "player") then
        f:RegisterEvent("PLAYER_ENTERING_WORLD")
        f:RegisterEvent("PLAYER_TARGET_CHANGED")
    end
    if(unit == "focus") then f:RegisterEvent("PLAYER_FOCUS_CHANGED") end
    
    if(unit == "player" or unit == "target" or unit == "pet" or unit == "focus") then
        f:RegisterEvent("UNIT_HEALTH")
        f:RegisterEvent("UNIT_AURA")
        
        f:RegisterEvent("DISPLAYPOWER")
        f:RegisterEvent("UNIT_RUNIC_POWER")
        f:RegisterEvent("UNIT_ENERGY")
        f:RegisterEvent("UNIT_FOCUS")
        f:RegisterEvent("UNIT_MANA")
        f:RegisterEvent("UNIT_RAGE")
        f:RegisterEvent("UNIT_MAXRUNIC_POWER")
        f:RegisterEvent("UNIT_MAXENERGY")
        f:RegisterEvent("UNIT_MAXFOCUS")
        f:RegisterEvent("UNIT_MAXMANA")
        f:RegisterEvent("UNIT_MAXRAGE")
        f:RegisterEvent("UNIT_THREAT_SITUATION_UPDATE")
        
        f:RegisterEvent("UNIT_NAME_UPDATE")
        
        if(unit == "target" or unit == "pet" or unit == "focus") then
            f:RegisterEvent("UNIT_TARGET")
        end
    elseif(f.isPartyFrame or f.isRaidFrame) then
        f:RegisterEvent("PARTY_MEMBERS_CHANGED")
        f:RegisterEvent("RAID_ROSTER_UPDATE")
    end
    
    f:RegisterEvent("RAID_TARGET_UPDATE")
    
    f:SetScript("OnEvent", function(self, e)
        if(not UnitExists(self.unit)) then return end
        
        if(e == "UNIT_TARGET") then
            if(arg1 == self.unit) then
                local i
                for i = 1, #self.ancestors do
                    atrUF:UpdateAuras(self.ancestors[i])
                    atrUF:UpdateName(self.ancestors[i])
                    atrUF:UpdateHealth(self.ancestors[i])
                    atrUF:UpdatePower(self.ancestors[i])
                    atrUF:UpdateThreat(self.ancestors[i])
                    atrUF:UpdateRaidIcon(self.ancestors[i])
                    atrUF:UpdateRange(self.ancestors[i])
                end
            end
        elseif(e == "UNIT_AURA") then
            if(arg1 == self.unit) then                
                local i
                for i = 1, #atrUF.units do
                    if(UnitIsUnit(atrUF.units[i], self.unit)) then
                        atrUF:UpdateAuras(atrUF.units[i])
                    end
                end
            end
        elseif(e == "UNIT_HEALTH") then
            if(arg1 == self.unit) then
                local i
                for i = 1, #atrUF.units do
                    if(UnitIsUnit(atrUF.units[i], self.unit)) then
                        atrUF:UpdateHealth(atrUF.units[i])
                    end
                end
            end
        elseif(e == "DISPLAYPOWER" or e == "UNIT_RUNIC_POWER" or e == "UNIT_ENERGY" or e == "UNIT_FOCUS" or e == "UNIT_MANA" or e == "UNIT_RAGE" or e == "UNIT_MAXRUNIC_POWER" or e == "UNIT_MAXENERGY" or e == "UNIT_MAXFOCUS" or e == "UNIT_MAXMANA" or e == "UNIT_MAXRAGE") then
            if(arg1 == self.unit) then
                local i
                for i = 1, #atrUF.units do
                    if(UnitIsUnit(atrUF.units[i], self.unit)) then
                        atrUF:UpdatePower(atrUF.units[i])
                    end
                end
            end
        elseif(e == "UNIT_NAME_UPDATE") then
            if(arg1 == self.unit) then
                local i
                for i = 1, #atrUF.units do
                    if(UnitIsUnit(atrUF.units[i], self.unit)) then
                        atrUF:UpdateName(atrUF.units[i])
                    end
                end
            end
        elseif(e == "UNIT_THREAT_SITUATION_UPDATE") then
            if(arg1 == self.unit) then
                local i
                for i = 1, #atrUF.units do
                    if(UnitIsUnit(atrUF.units[i], self.unit)) then
                        atrUF:UpdateThreat(atrUF.units[i])
                    end
                end
            end
        elseif(e == "RAID_TARGET_UPDATE") then
            atrUF:UpdateRaidIcon(self.unit)
        elseif(e == "PLAYER_ENTERING_WORLD") then
            atrUF:UpdateHealth("player")
            atrUF:UpdatePower("player")
        elseif(e == "PLAYER_TARGET_CHANGED") then
            atrUF:UpdateAuras("target")
            atrUF:UpdateName("target")
            atrUF:UpdateHealth("target")
            atrUF:UpdatePower("target")
            atrUF:UpdateThreat("player")
            atrUF:UpdateThreat("target")
            atrUF:UpdateLevel("target")
            atrUF:UpdateRaidIcon("target")
            atrUF:UpdateRange("target")
            
            local i
            for i = 1, #atrUF.target.ancestors do
                atrUF:UpdateAuras(atrUF.target.ancestors[i])
                atrUF:UpdateName(atrUF.target.ancestors[i])
                atrUF:UpdateHealth(atrUF.target.ancestors[i])
                atrUF:UpdatePower(atrUF.target.ancestors[i])
                atrUF:UpdateRaidIcon(atrUF.target.ancestors[i])
                atrUF:UpdateRange(atrUF.target.ancestors[i])
            end
        elseif(e == "PLAYER_FOCUS_CHANGED") then
            atrUF:UpdateAuras("focus")
            atrUF:UpdateName("focus")
            atrUF:UpdateHealth("focus")
            atrUF:UpdatePower("focus")
            atrUF:UpdateThreat("focus")
            atrUF:UpdateRaidIcon("focus")
            atrUF:UpdateRange("focus")
            
            local i
            for i = 1, #self.ancestors do
                atrUF:UpdateAuras(self.ancestors[i])
                atrUF:UpdateName(self.ancestors[i])
                atrUF:UpdateHealth(self.ancestors[i])
                atrUF:UpdatePower(self.ancestors[i])
                atrUF:UpdateThreat(self.ancestors[i])
                atrUF:UpdateRange(self.ancestors[i])
            end
        elseif(e == "PARTY_MEMBERS_CHANGED") then
            atrUF:UpdateAuras(self.unit)
            atrUF:UpdateName(self.unit)
            atrUF:UpdateHealth(self.unit)
            atrUF:UpdatePower(self.unit)
            atrUF:UpdateThreat(self.unit)
            atrUF:UpdateRaidIcon(self.unit)
            atrUF:UpdateRange(self.unit)
        elseif(e == "RAID_ROSTER_UPDATE") then
            atrUF:UpdateAuras(self.unit)
            atrUF:UpdateName(self.unit)
            atrUF:UpdateHealth(self.unit)
            atrUF:UpdatePower(self.unit)
            atrUF:UpdateThreat(self.unit)
            atrUF:UpdateRaidIcon(self.unit)
            atrUF:UpdateRange(self.unit)
        end
    end)
    
    f:SetScript("OnShow", function(self)
        atrUF:UpdateAuras(self.unit)
        atrUF:UpdateName(self.unit)
        atrUF:UpdateHealth(self.unit)
        atrUF:UpdatePower(self.unit)
        atrUF:UpdateRaidIcon(self.unit)
        atrUF:UpdateRange(self.unit)
    end)
    f:SetScript("OnEnter", function(self)
        GameTooltip_SetDefaultAnchor(GameTooltip, self)
        GameTooltip:SetUnit(self.unit)
        
        local r, g, b = GameTooltip_UnitColor(self.unit)
        GameTooltipTextLeft1:SetTextColor(r, g, b)
    end)
    f:SetScript("OnLeave", function(self)
        GameTooltip:Hide()
    end)
    
    if(unit ~= "player") then
        f.nextUpdate = 1
        f:SetScript("OnUpdate", function(self, elapsed)            
            if(self.nextUpdate <= .5) then
                self.nextUpdate = 1
                atrUF:UpdateRange(self.unit)
            else
                self.nextUpdate = self.nextUpdate-elapsed
            end
        end)
    end
    
    ------------------------------------
        
    return f
end

function atrUF:BuildAncestry()
    local i
    for i = 1, #self.units do
        self[self.units[i]].ancestors = {}
        
        local u
        for u = 1, #self.units do
            if(self.units[i] ~= self.units[u]) then
                if(string.find(self.units[u], "^("..self.units[i]..")")) then
                    table.insert(self[self.units[i]].ancestors, self.units[u])
                end
            end
        end
    end
end